# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The base time-series array object.

This module defines the following classes

--------------------  ---------------------------------------------------------
`TimeSeriesBase`      base of the `TimeSeries` and `StateVector` classes,
                      provides the constructor and all common methods
                      (mainly everything that isn't signal-processing related)
`TimeSeriesBaseDict`  base of the `TimeSeriesDict`, this exists mainly so that
                      the `TimeSeriesDict` and `StateVectorDict` can be
                      distinct objects
`TimeSeriesBaseList`  base of the `TimeSeriesList` and `StateVectorList`,
                      same reason for living as the `TimeSeriesBaseDict`
--------------------  ---------------------------------------------------------

**None of these objects are really designed to be used other than as bases for
user-facing objects.**
"""

from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    Generic,
    TypeVar,
    cast,
    overload,
)

import numpy
from astropy import units
from astropy.units import Quantity
from gwosc.api import DEFAULT_URL as GWOSC_DEFAULT_HOST

from ..detector import Channel
from ..io.registry import UnifiedReadWriteMethod
from ..segments import SegmentList
from ..time import (
    LIGOTimeGPS,
    Time,
    to_gps,
)
from ..types import Series
from ..utils.misc import property_alias
from .connect import (
    TimeSeriesBaseDictGet,
    TimeSeriesBaseDictRead,
    TimeSeriesBaseDictWrite,
    TimeSeriesBaseGet,
    TimeSeriesBaseRead,
    TimeSeriesBaseWrite,
)

if TYPE_CHECKING:
    import re
    from collections.abc import (
        Callable,
        Iterable,
        Mapping,
    )
    from typing import (
        ClassVar,
        Literal,
        Self,
        SupportsFloat,
        SupportsIndex,
    )

    import arrakis
    import nds2
    import pycbc
    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from numpy.typing import (
        DTypeLike,
        NDArray,
    )

    from ..plot import Plot
    from ..segments import Segment
    from ..time import SupportsToGps
    from ..typing import (
        ArrayLike1D,
        UnitLike,
    )
    from ..utils.lal import LALTimeSeriesType
    from .statevector import StateTimeSeries

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "TimeSeriesBase",
    "TimeSeriesBaseDict",
    "TimeSeriesBaseList",
]


_UFUNC_STRING = {
    "less": "<",
    "less_equal": "<=",
    "equal": "==",
    "greater_equal": ">=",
    "greater": ">",
}


# -- utilities -----------------------

def _format_time(gps: Time | Quantity | SupportsFloat) -> float:
    """Format a GPS time into a float."""
    if isinstance(gps, Time):
        return gps.gps
    if isinstance(gps, Quantity):
        return gps.to(units.second).value
    return float(gps)


def _dynamic_scaled(
    scaled: bool | None,  # noqa: FBT001
    channel: str | Channel,
) -> bool:
    """Determine default for scaled based on channel name.

    This is mainly to work around LIGO not correctly recording ADC
    scaling parameters for most of Advanced LIGO (through 2023).
    Scaling parameters for H0 and L0 data are also not correct
    starting in mid-2020.

    Parameters
    ----------
    scaled : `bool`, `None`
        The scaled argument as given by the user.

    channel : `str`, `Channel`
        The name of the channel to be read.

    Returns
    -------
    scaled : `bool`
        `False` if channel is from LIGO, otherwise `True`.

    Examples
    --------
    >>> _dynamic_scaled(None, "H1:channel")
    False
    >>> _dynamic_scaled(None, "V1:channel")
    True
    """
    if scaled is not None:
        return scaled
    return not str(channel).startswith(("H0", "L0", "H1", "L1"))


# -- TimeSeriesBase-------------------

class TimeSeriesBase(Series):
    """An `Array` with time-domain metadata.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`, optional
        Physical unit of these data.

    t0 : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    dt : `float`, `~astropy.units.Quantity`, optional
        Time between successive samples (seconds), can also be given inversely
        via `sample_rate`.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        The rate of samples per second (Hertz), can also be given inversely
        via `dt`.

    times : `array-like`
        The complete array of GPS times accompanying the data for this series.
        This argument takes precedence over `t0` and `dt` so should be given
        in place of these if relevant, not alongside.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional, default: `False`
        Choose to copy the input data to new memory.

    subok : `bool`, optional, default: `True`
        Allow passing of sub-classes by the array generator.
    """

    _default_xunit: ClassVar[UnitBase] = units.second
    _print_slots: ClassVar[tuple[str, ...]] = (
        "t0",
        "dt",
        "name",
        "channel",
    )
    DictClass: ClassVar[type[TimeSeriesBaseDict]]

    def __new__(
        cls,
        data: ArrayLike1D,
        unit: UnitLike = None,
        t0: SupportsToGps | None = None,
        dt: float | Quantity | None = None,
        sample_rate: float | Quantity | None = None,
        times: ArrayLike1D | None = None,
        channel: Channel | str | None = None,
        name: str | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new `TimeSeriesBase`."""
        # parse t0 or epoch
        epoch = kwargs.pop("epoch", None)
        if epoch is not None and t0 is not None:
            msg = "give only one of epoch or t0"
            raise ValueError(msg)
        if epoch is None and t0 is not None:
            kwargs["x0"] = _format_time(t0)
        elif epoch is not None:
            kwargs["x0"] = _format_time(epoch)

        # parse sample_rate or dt
        if sample_rate is not None and dt is not None:
            msg = "give only one of sample_rate or dt"
            raise ValueError(msg)
        if sample_rate is None and dt is not None:
            kwargs["dx"] = dt

        # parse times
        if times is not None:
            kwargs["xindex"] = times

        # generate TimeSeries
        new = super().__new__(
            cls,
            data,
            name=name,
            unit=unit,
            channel=channel,
            **kwargs,
        )

        # manually set sample_rate if given
        if sample_rate is not None:
            new.sample_rate = sample_rate

        return new

    # -- TimeSeries properties -------

    # rename properties from the Series
    t0 = property_alias(Series.x0, "GPS start time of this series.")  # type: ignore[arg-type]
    dt = property_alias(Series.dx, "Time (seconds) between successive samples.")  # type: ignore[arg-type]
    span = property_alias(Series.xspan, "Time (seconds) spanned by this series.")  # type: ignore[arg-type]
    times = property_alias(Series.xindex, "Array of GPS times for each sample.")  # type: ignore[arg-type]

    # -- epoch
    # this gets redefined to attach to the t0 property
    @property
    def epoch(self) -> Time | None:
        """GPS epoch for these data.

        This attribute is stored internally by the `t0` attribute.
        """
        try:
            return Time(self.t0, format="gps", scale="utc")
        except AttributeError:
            return None

    @epoch.setter
    def epoch(self, epoch: Time | SupportsToGps | None) -> None:
        if epoch is None:
            del self.t0
        elif isinstance(epoch, Time):
            self.t0 = epoch.gps
        else:
            try:
                self.t0 = to_gps(epoch)  # type: ignore[assignment]
            except TypeError:
                self.t0 = epoch

    # -- sample_rate
    @property
    def sample_rate(self) -> Quantity:
        """Data rate for this `TimeSeries` in samples per second (Hertz).

        This attribute is stored internally by the `dx` attribute
        """
        return (1 / self.dt).to("Hertz")

    @sample_rate.setter
    def sample_rate(self, val: QuantityLike | None) -> None:
        if val is None:
            del self.dt
            return
        self.dt = (1 / Quantity(val, units.Hertz)).to(self.xunit)

    # -- duration
    @property
    def duration(self) -> Quantity:
        """Duration of this series in seconds.

        :type: `~astropy.units.Quantity` scalar
        """
        return Quantity(
            abs(self.span),
            self.xunit,
            dtype=float,
        )

    # -- TimeSeries i/o --------------

    read = UnifiedReadWriteMethod(TimeSeriesBaseRead)
    write = UnifiedReadWriteMethod(TimeSeriesBaseWrite)
    get = UnifiedReadWriteMethod(TimeSeriesBaseGet)

    # -- TimeSeries accessors --------

    @classmethod
    def fetch(
        cls,
        channel: str | Channel,
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        host: str | None = None,
        port: int | None = None,
        verbose: bool | str = False,
        connection: nds2.connection | None = None,
        verify: bool = False,
        pad: float | None = None,
        allow_tape: bool | None = None,
        scaled: bool | None = None,
        type: int | str | None = None,  # noqa: A002
        dtype: int | str | None = None,
    ) -> Self:
        """Fetch data from NDS.

        Parameters
        ----------
        channel : `str`, `~gwpy.detector.Channel`
            The name (or representation) of the data channel to fetch.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        host : `str`, optional
            URL of NDS server to use, if blank will try any server
            (in a relatively sensible order) to get the data

            One of ``connection`` or ``host`` must be given.

        port : `int`, optional
            Port number for NDS server query, must be given with `host`.

        verify : `bool`, optional
            Check channels exist in database before asking for data.
            Default is `True`.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        connection : `nds2.connection`, optional
            Open NDS connection to use.
            Default is to open a new connection using ``host`` and ``port``
            arguments.

            One of ``connection`` or ``host`` must be given.

        pad : `float`, optional
            Float value to insert between gaps.
            Default behaviour is to raise an exception when any gaps are
            found.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        allow_tape : `bool`, optional
            Allow data access from slow tapes.
            If ``host`` or ``connection`` is given, the default is to do
            whatever the server default is, otherwise servers will be searched
            with ``allow_tape=False`` first, then ``allow_tape=True`` if that
            fails.

        type : `int`, `str`, optional
            NDS2 channel type integer or string name to match.
            Default is to search for any channel type.

        dtype : `numpy.dtype`, `str`, `type`, or `dict`, optional
            NDS2 data type to match.
            Default is to search for any data type.
        """
        return cls.get(
            channel,
            start,
            end,
            source="nds2",
            host=host,
            port=port,
            verbose=verbose,
            connection=connection,
            verify=verify,
            pad=pad,
            scaled=scaled,
            allow_tape=allow_tape,
            type=type,
            dtype=dtype,
        )

    @classmethod
    def fetch_open_data(
        cls,
        ifo: str,
        start: SupportsToGps,
        end: SupportsToGps,
        sample_rate: float = 4096,
        version: int | None = None,
        format: Literal["gwf", "hdf5"] = "hdf5",  # noqa: A002
        host: str = GWOSC_DEFAULT_HOST,
        *,
        verbose: bool | None = None,
        cache: bool | None = None,
        **kwargs,
    ) -> Self:
        """Fetch open-access data from GWOSC.

        This is just a shim around ``TimeSeries.get(..., source='gwosc')``.

        Parameters
        ----------
        ifo : `str`
            The two-character prefix of the IFO in which you are interested,
            e.g. `'L1'`.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS start time of required data, defaults to start of data found;
            any input parseable by `~gwpy.time.to_gps` is fine.

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine.

        sample_rate : `float`, optional,
            The sample rate of desired data; most data are stored
            by GWOSC at 4096 Hz, however there may be event-related
            data releases with a 16384 Hz rate, default: `4096`.

        version : `int`, optional
            Version of files to download, defaults to highest discovered
            version.

        format : `str`, optional
            The data format to download and parse, default: ``'h5py'``

            - ``'hdf5'``
            - ``'gwf'`` - requires |lalframe|_

        host : `str`, optional
            HTTP host name of GWOSC server to access.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        cache : `bool`, optional
            Save/read a local copy of the remote URL, default: `False`;
            useful if the same remote data are to be accessed multiple times.
            Set ``GWPY_CACHE=1`` in the environment to auto-cache.

        timeout : `float`, optional
            The time to wait for a response from the GWOSC server.

        kwargs
            Any other keyword arguments are passed to the `TimeSeries.read`
            method that parses the file that was downloaded.

        Examples
        --------
        >>> from gwpy.timeseries import (TimeSeries, StateVector)
        >>> print(TimeSeries.fetch_open_data('H1', 1126259446, 1126259478))
        TimeSeries([  2.17704028e-19,  2.08763900e-19,  2.39681183e-19,
                    ...,   3.55365541e-20,  6.33533516e-20,
                      7.58121195e-20]
                   unit: Unit(dimensionless),
                   t0: 1126259446.0 s,
                   dt: 0.000244140625 s,
                   name: Strain,
                   channel: None)
        >>> print(StateVector.fetch_open_data('H1', 1126259446, 1126259478))
        StateVector([127,127,127,127,127,127,127,127,127,127,127,127,
                     127,127,127,127,127,127,127,127,127,127,127,127,
                     127,127,127,127,127,127,127,127]
                    unit: Unit(dimensionless),
                    t0: 1126259446.0 s,
                    dt: 1.0 s,
                    name: quality/simple,
                    channel: None,
                    bits: Bits(0: data present
                               1: passes cbc CAT1 test
                               2: passes cbc CAT2 test
                               3: passes cbc CAT3 test
                               4: passes burst CAT1 test
                               5: passes burst CAT2 test
                               6: passes burst CAT3 test,
                               channel=None,
                               epoch=1126259446.0))

        For the `StateVector`, the naming of the bits will be
        ``format``-dependent, because they are recorded differently by GWOSC
        in different formats.

        Notes
        -----
        `StateVector` data are not available in ``txt.gz`` format.
        """
        return cls.get(
            ifo,
            start,
            end,
            source="gwosc",
            sample_rate=sample_rate,
            version=version,
            format=format,
            verbose=verbose,
            cache=cache,
            host=host,
            series_class=cls,
            **kwargs,
        )

    @classmethod
    def find(
        cls,
        channel: str | Channel,
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        observatory: str | None = None,
        frametype: str | None = None,
        frametype_match: str | re.Pattern | None = None,
        host: str | None = None,
        urltype: str | None = "file",
        ext: str = "gwf",
        pad: float | None = None,
        scaled: bool | None = None,
        allow_tape: bool | None = None,
        parallel: int = 1,
        verbose: bool | str = False,
        **readargs,
    ) -> Self:
        """Find and return data for multiple channels using GWDataFind.

        This method uses :mod:`gwdatafind` to discover the URLs
        that provide the requested data, then reads those files using
        :meth:`TimeSeriesDict.read()`.

        This is just a shim around ``TimeSeries.get(..., source='gwdatafind')``.

        Parameters
        ----------
        channel : `str`
            Name of data channel to find.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine.

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine.

        observatory : `str`, optional
            The observatory to use when searching for data.
            Default is to use the observatory from the channel name prefix,
            but this should be specified when searching for data in a
            multi-observatory dataset (e.g. `observatory='HLV'`).

        frametype : `str`, optional
            Name of frametype (dataset) in which this channel is stored.
            Default is to search all available datasets for a match, which
            can be very slow.

        frametype_match : `str`, optional
            Regular expression to use for frametype matching.

        host : `str`, optional
            Name of the GWDataFind server to use.
            Default is set by `gwdatafind.utils.get_default_host`.

        urltype : `str`, optional
            The URL type to use.
            Default is "file" to use paths available on the file system.

        ext : `str`, optional
            The file extension for which to search.
            "gwf" is the only file extension supported, but this may be
            extended in the future.

        pad : `float`, optional
            Value with which to fill gaps in the source data,
            by default gaps will result in a `ValueError`.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        parallel : `int`, optional
            Number of parallel processes to use.

        allow_tape : `bool`, optional
            Allow reading from frame files on (slow) magnetic tape.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        readargs
            Any other keyword arguments to be passed to `.read()`.
        """
        return cls.get(
            channel,
            start,
            end,
            source="gwdatafind",
            observatory=observatory,
            frametype=frametype,
            frametype_match=frametype_match,
            host=host,
            urltype=urltype,
            ext=ext,
            verbose=verbose,
            pad=pad,
            scaled=scaled,
            allow_tape=allow_tape,
            parallel=parallel,
            **readargs,
        )

    # -- utilities -------------------

    def plot(
        self,
        method: str = "plot",
        figsize: tuple[int, int] = (12, 4),
        xscale: str = "auto-gps",
        **kwargs,
    ) -> Plot:
        """Plot the data for this timeseries.

        Returns
        -------
        figure : `~matplotlib.figure.Figure`
            The newly created figure, with populated Axes.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the
            figure.
        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the
            axes.
        matplotlib.axes.Axes.plot
            For documentation of keyword arguments used in rendering the data.
        """
        kwargs.update(figsize=figsize, xscale=xscale)
        return super().plot(method=method, **kwargs)

    @classmethod
    def from_arrakis(
        cls,
        series: arrakis.block.Series,
        *,
        copy: bool = True,
        **metadata,
    ) -> Self:
        """Construct a new series from an `arrakis.Series` object.

        Parameters
        ----------
        series : `arrakis.Series`
            The input Arrakis data series to read.

        copy : `bool`, optional
            If `True`, copy the contained data array to new to a new array.

        metadata
            Any other metadata keyword arguments to pass to the `TimeSeries`
            constructor.

        Returns
        -------
        timeseries : `TimeSeries`
            A new `TimeSeries` containing the data from the `arrakis.Series`
            and the appropriate metadata.
        """
        # get Channel from buffer
        channel = Channel.from_arrakis(series.channel)

        # set default metadata
        defaults = {
            "channel": channel,
            "epoch": LIGOTimeGPS(0, series.time_ns),
            "dt": series.dt,
            "unit": None,
            "name": series.name,
        }
        metadata = {**defaults, **metadata}

        # construct new TimeSeries-like object
        return cls(series.data, copy=copy, **metadata)

    @classmethod
    def from_nds2_buffer(
        cls,
        buffer: nds2.buffer,
        *,
        scaled: bool | None = None,
        copy: bool = True,
        **metadata,
    ) -> Self:
        """Construct a new series from an `nds2.buffer` object.

        **Requires:** |nds2|_

        Parameters
        ----------
        buffer : `nds2.buffer`
            The input NDS2-client buffer to read.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        copy : `bool`, optional
            Tf `True`, copy the contained data array to new  to a new array.

        metadata
            Any other metadata keyword arguments to pass to the `TimeSeries`
            constructor.

        Returns
        -------
        timeseries : `TimeSeries`
            A new `TimeSeries` containing the data from the `nds2.buffer`,
            and the appropriate metadata.
        """
        # get Channel from buffer
        channel = Channel.from_nds2(buffer.channel)

        # set default metadata
        defaults = {
            "channel": channel,
            "epoch": LIGOTimeGPS(buffer.gps_seconds, buffer.gps_nanoseconds),
            "sample_rate": channel.sample_rate,
            "unit": channel.unit,
            "name": buffer.name,
        }
        metadata = {**defaults, **metadata}

        # unwrap data
        scaled = _dynamic_scaled(scaled, channel.name)
        slope = buffer.signal_slope
        offset = buffer.signal_offset
        null_scaling = slope == 1. and offset == 0.
        if scaled and not null_scaling:
            data = buffer.data.copy() * slope + offset
            copy = False
        else:
            data = buffer.data

        # construct new TimeSeries-like object
        return cls(data, copy=copy, **metadata)

    @classmethod
    def from_lal(
        cls,
        lalts: LALTimeSeriesType,
        *,
        copy: bool = True,
    ) -> Self:
        """Generate a new TimeSeries from a LAL TimeSeries of any type."""
        # convert the units
        from ..utils.lal import (
            from_lal_type,
            from_lal_unit,
        )

        unit = from_lal_unit(lalts.sampleUnits)

        dtype: DTypeLike
        try:
            dtype = lalts.data.data.dtype
        except AttributeError:  # no data
            dtype = from_lal_type(type(lalts))
            data = numpy.array([], dtype=dtype)
        else:
            data = lalts.data.data

        # create new series
        out = cls(
            data,
            dtype=dtype,
            name=lalts.name or None,
            unit=unit,
            t0=lalts.epoch,
            dt=lalts.deltaT,
            channel=None,
            copy=False,
        )

        if copy:
            return out.copy()
        return out

    def to_lal(self) -> LALTimeSeriesType:
        """Convert this `TimeSeries` into a LAL TimeSeries.

        .. note::

           This operation always copies data to new memory.
        """
        import lal

        from ..utils.lal import (
            find_typed_function,
            to_lal_unit,
        )

        # map unit
        try:
            unit, scale = to_lal_unit(self.unit)
        except ValueError as exc:
            warnings.warn(
                f"{exc}, defaulting to lal.DimensionlessUnit",
                stacklevel=2,
            )
            unit = lal.DimensionlessUnit
            scale = 1

        # create TimeSeries
        create = find_typed_function(self.dtype, "Create", "TimeSeries")
        lalts = create(
            self.name or str(self.channel or "") or None,
            LIGOTimeGPS(self.t0.value),
            0,
            self.dt.value,
            unit,
            self.shape[0],
        )

        # assign data
        lalts.data.data = self.value
        if scale != 1:
            lalts.data.data *= scale

        return lalts

    @classmethod
    def from_pycbc(
        cls,
        pycbcseries: pycbc.types.TimeSeries,
        *,
        copy: bool = True,
    ) -> Self:
        """Convert a `pycbc.types.timeseries.TimeSeries` into a `TimeSeries`.

        Parameters
        ----------
        pycbcseries : `pycbc.types.timeseries.TimeSeries`
            The input PyCBC `~pycbc.types.timeseries.TimeSeries` array.

        copy : `bool`, optional
            If `True`, copy these data to a new array.

        Returns
        -------
        timeseries : `TimeSeries`
            A GWpy version of the input timeseries.
        """
        return cls(
            pycbcseries.data,
            t0=pycbcseries.start_time,
            dt=pycbcseries.delta_t,
            copy=copy,
        )

    def to_pycbc(self, *, copy: bool = True) -> pycbc.types.TimeSeries:
        """Convert this `TimeSeries` into a PyCBC `~pycbc.types.timeseries.TimeSeries`.

        Parameters
        ----------
        copy : `bool`, optional, default: `True`
            If `True`, copy these data to a new array.

        Returns
        -------
        timeseries : `~pycbc.types.timeseries.TimeSeries`
            A PyCBC representation of this `TimeSeries`.
        """
        from pycbc import types
        return types.TimeSeries(
            self.value,
            delta_t=self.dt.to("s").value,
            epoch=self.t0.value, copy=copy,
        )

    # -- TimeSeries operations -------

    def __array_ufunc__(  # type: ignore[override]
        self,
        function: Callable,
        method: str,
        *inputs,
        **kwargs,
    ) -> Self | Quantity | StateTimeSeries:
        """Override the default array ufunc to handle TimeSeries metadata."""
        out = super().__array_ufunc__(function, method, *inputs, **kwargs)
        if out.dtype is numpy.dtype(bool) and len(inputs) == 2:
            from .statevector import StateTimeSeries
            orig, value = inputs
            try:
                op_ = _UFUNC_STRING[function.__name__]
            except KeyError:
                op_ = function.__name__
            out = out.view(StateTimeSeries)
            out.__metadata_finalize__(orig)
            oname = orig.name if isinstance(orig, type(self)) else orig
            vname = value.name if isinstance(value, type(self)) else value
            out.name = f"{oname!s} {op_!s} {vname!s}"
        return out

    # Quantity overrides __eq__ and __ne__ in a way that doesn't work for us,
    # so we just undo that
    def __eq__(self, other: object) -> bool:
        """Return `True` if ``other`` is equal to this `TimeSeries`."""
        return numpy.ndarray.__eq__(self, other)

    def __ne__(self, other: object) -> bool:
        """Return `True` if ``other`` is not equal to this `TimeSeries`."""
        return numpy.ndarray.__ne__(self, other)


# -- TimeSeriesBaseDict --------------

def as_series_dict_class(
    seriesclass: type[TimeSeriesBase],
) -> Callable[[type[TimeSeriesBaseDict]], type[TimeSeriesBaseDict]]:
    """Return a decorator for a `dict` class to define `DictClass` for its `EntryClass`.

    This method should be used to decorate sub-classes of the
    `TimeSeriesBaseDict` to provide a reference to that class from the
    relevant subclass of `TimeSeriesBase`.
    """
    def decorate_class(cls: type[TimeSeriesBaseDict]) -> type[TimeSeriesBaseDict]:
        """Set ``cls`` as the `DictClass` attribute for this series type."""
        seriesclass.DictClass = cls
        return cls

    return decorate_class


# Type variable for generic dict values
_V = TypeVar("_V", bound=TimeSeriesBase)


@as_series_dict_class(TimeSeriesBase)
class TimeSeriesBaseDict(dict[str | Channel, _V], Generic[_V]):
    """Key-value mapping of named `TimeSeriesBase` objects.

    This object is designed to hold data for many different sources (channels)
    for a single time span. Dictionary keys are ordered by insertion order.

    The main entry points for this object are the
    :meth:`~TimeSeriesBaseDict.read` and :meth:`~TimeSeriesBaseDict.fetch`
    data access methods.
    """

    EntryClass: ClassVar[type[TimeSeriesBase]] = TimeSeriesBase

    @property
    def span(self) -> Segment:
        """The GPS ``[start, stop)`` extent of data in this `dict`."""
        span = SegmentList()
        for value in self.values():
            span.append(value.span)
        try:
            return span.extent()
        except ValueError as exc:  # empty list
            exc.args = (
                f"cannot calculate span for empty {type(self).__name__}",
            )
            raise

    read = UnifiedReadWriteMethod(TimeSeriesBaseDictRead)
    write = UnifiedReadWriteMethod(TimeSeriesBaseDictWrite)
    get = UnifiedReadWriteMethod(TimeSeriesBaseDictGet)

    def __iadd__(self, other: dict[str | Channel, numpy.ndarray]) -> Self:
        """Append a `TimeSeriesBase` or `numpy.ndarray` to this dict."""
        self.append(other)
        return self

    def copy(self) -> Self:
        """Return a copy of this dict with each value copied to new memory."""
        new = self.__class__()
        for key, val in self.items():
            new[key] = val.copy()
        return new

    def append(
        self,
        other: Mapping[str | Channel, NDArray],
        *,
        copy: bool = True,
        **kwargs,
    ) -> Self:
        """Append the dict ``other`` to this one.

        Parameters
        ----------
        other : `dict` of `TimeSeries`
            The container to append to this one.

        copy : `bool`, optional
            If `True` copy data from ``other`` before storing, only
            affects those keys in ``other`` that aren't in ``self``.

        **kwargs
            Other keyword arguments to send to `TimeSeries.append`.

        See Also
        --------
        TimeSeries.append
            For details of the underlying series append operation.
        """
        for key, series in other.items():
            if key in self:
                self[key].append(series, **kwargs)
            else:
                if not isinstance(series, self.EntryClass):
                    msg = (
                        f"cannot append {type(series).__name__} to "
                        f"{type(self).__name__} for new key {key!r}"
                    )
                    raise ValueError(msg)
                if copy:
                    series = series.copy()  # noqa: PLW2901
                self[key] = series
        return self

    def prepend(
        self,
        other: Mapping[str | Channel, _V],
        **kwargs,
    ) -> Self:
        """Prepend the dict ``other`` to this one.

        Parameters
        ----------
        other : `dict` of `TimeSeries`
            The container to prepend to this one.

        copy : `bool`, optional
            If `True` copy data from ``other`` before storing, only
            affects those keys in ``other`` that aren't in ``self``.

        kwargs
            Other keyword arguments to send to `TimeSeries.prepend`.

        See Also
        --------
        TimeSeries.prepend
            For details of the underlying series prepend operation.
        """
        for key, series in other.items():
            if key in self:
                self[key].prepend(series, **kwargs)
            else:
                self[key] = series
        return self

    def crop(
        self,
        start: SupportsToGps | None = None,
        end: SupportsToGps | None = None,
        *,
        copy: bool = False,
    ) -> Self:
        """Crop each entry of this `dict`.

        This method calls the :meth:`crop` method of all entries and
        modifies this dict in place.

        Parameters
        ----------
        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        copy : `bool`, optional, default: `False`
            If `True` copy the data for each entry to fresh memory,
            otherwise return a view.

        See Also
        --------
        TimeSeries.crop
            for more details
        """
        for key, val in self.items():
            self[key] = val.crop(start=start, end=end, copy=copy)
        return self

    def resample(
        self,
        rate: dict[str | Channel, float] | float,
        **kwargs,
    ) -> Self:
        """Resample items in this dict.

        This operation over-writes items inplace.

        Parameters
        ----------
        rate : `dict`, `float`
            either a `dict` of (channel, `float`) pairs for key-wise
            resampling, or a single float/int to resample all items.

        **kwargs
             other keyword arguments to pass to each item's resampling
             method.
        """
        if not isinstance(rate, dict):
            rate = dict.fromkeys(self, rate)
        for key, resamp in rate.items():
            self[key] = self[key].resample(resamp, **kwargs)
        return self

    @classmethod
    def fetch(
        cls,
        channels: list[str | Channel],
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        host: str | None = None,
        port: int | None = None,
        verbose: bool | str = False,
        connection: nds2.connection | None = None,
        verify: bool = False,
        pad: float | None = None,
        allow_tape: bool | None = None,
        scaled: bool | None = None,
        type: int | str | None = None,  # noqa: A002
        dtype: int | str | None = None,
    ) -> Self:
        """Fetch data from NDS for a number of channels.

        This is just a shim around ``TimeSeriesDict.get(..., source='nds2')``.

        Parameters
        ----------
        channels : `str`, `~gwpy.detector.Channel`
            List of names of data channels to find.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        host : `str`, optional
            URL of NDS server to use, if blank will try any server
            (in a relatively sensible order) to get the data

            One of ``connection`` or ``host`` must be given.

        port : `int`, optional
            Port number for NDS server query, must be given with `host`.

        verify : `bool`, optional
            Check channels exist in database before asking for data.
            Default is `True`.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        connection : `nds2.connection`, optional
            Open NDS connection to use.
            Default is to open a new connection using ``host`` and ``port``
            arguments.

            One of ``connection`` or ``host`` must be given.

        pad : `float`, optional
            Float value to insert between gaps.
            Default behaviour is to raise an exception when any gaps are
            found.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        allow_tape : `bool`, optional
            Allow data access from slow tapes.
            If ``host`` or ``connection`` is given, the default is to do
            whatever the server default is, otherwise servers will be searched
            with ``allow_tape=False`` first, then ``allow_tape=True`` if that
            fails.

        type : `int`, `str`, optional
            NDS2 channel type integer or string name to match.
            Default is to search for any channel type.

        dtype : `numpy.dtype`, `str`, `type`, or `dict`, optional
            NDS2 data type to match.
            Default is to search for any data type.

        Returns
        -------
        data : `TimeSeriesBaseDict`
            A new `TimeSeriesBaseDict` of (`str`, `TimeSeries`) pairs fetched
            from NDS.
        """
        return cls.get(
            channels,
            start,
            end,
            source="nds2",
            host=host,
            port=port,
            verify=verify,
            verbose=verbose,
            connection=connection,
            pad=pad,
            scaled=scaled,
            allow_tape=allow_tape,
            type=type,
            dtype=dtype,
            series_class=cls.EntryClass,
        )

    @classmethod
    def fetch_open_data(
        cls,
        detectors: str,
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        sample_rate: float = 4096,
        version: int | None = None,
        format: str = "hdf5",  # noqa: A002
        host: str = GWOSC_DEFAULT_HOST,
        verbose: bool | None = None,
        cache: bool | None = None,
        parallel: int = 1,
        **kwargs,
    ) -> Self:
        """Fetch open-access data from the LIGO Open Science Center.

        This is just a shim around ``TimeSeriesDict.get(..., source='gwosc')``.

        Parameters
        ----------
        detectors : `list` of `str`
            List of two-character prefices of the IFOs in which you
            are interested, e.g. `['H1', 'L1']`.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine.

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine.

        sample_rate : `float`, `Quantity`,
            The sample rate (Hertz) of desired data; most data are stored
            by GWOSC at 4096 Hz, however there may be event-related
            data releases with a 16384 Hz rate.

        version : `int`
            Version of files to download, defaults to highest discovered
            version.

        format : `str`
            The data format to download and parse.
            One of

            "hdf5"
                HDF5 data files, read using `h5py`.

            "gwf"
                Gravitational-Wave Frame files, requires |LDAStools.frameCPP|_.

        host : `str`
            Host name of GWOSC server to access.

        verbose : `bool`
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        cache : `bool`
            Save/read a local copy of the remote URL, default: `False`;
            useful if the same remote data are to be accessed multiple times.
            Set `GWPY_CACHE=1` in the environment to auto-cache.

        parallel : `int`
            Number of parallel threads to use when downloading data for
            multiple detectors. Default is ``1``.

        kwargs
            Any other keyword arguments are passed to the `TimeSeries.read`
            method that parses the file that was downloaded.

        See Also
        --------
        TimeSeries.fetch_open_data
            For more examples.

        TimeSeries.read
            For details of how files are read.

        Examples
        --------
        >>> from gwpy.timeseries import TimeSeriesDict
        >>> print(TimeSeriesDict.fetch_open_data(['H1', 'L1'], 1126259446, 1126259478))
        TimeSeriesDict({'H1': <TimeSeries([2.17704028e-19, 2.08763900e-19, 2.39681183e-19, ...,
                     3.55365541e-20, 6.33533516e-20, 7.58121195e-20]
                    unit=Unit(dimensionless),
                    t0=<Quantity 1.12625945e+09 s>,
                    dt=<Quantity 0.00024414 s>,
                    name='Strain',
                    channel=None)>, 'L1': <TimeSeries([-1.04289994e-18, -1.03586274e-18, -9.89322445e-19,
                     ..., -1.01767748e-18, -9.82876816e-19,
                     -9.59276974e-19]
                    unit=Unit(dimensionless),
                    t0=<Quantity 1.12625945e+09 s>,
                    dt=<Quantity 0.00024414 s>,
                    name='Strain',
                    channel=None)>})
        """  # noqa: E501
        return cls.get(
            detectors,
            start,
            end,
            source="gwosc",
            sample_rate=sample_rate,
            version=version,
            format=format,
            host=host,
            verbose=verbose,
            cache=cache,
            parallel=parallel,
            series_class=cls.EntryClass,
            **kwargs,
        )

    @classmethod
    def find(
        cls,
        channels: list[str | Channel],
        start: SupportsToGps,
        end: SupportsToGps,
        *,
        observatory: str | None = None,
        frametype: str | None = None,
        frametype_match: str | re.Pattern | None = None,
        host: str | None = None,
        urltype: str | None = "file",
        ext: str = "gwf",
        pad: float | None = None,
        scaled: bool | None = None,
        allow_tape: bool | None = None,
        parallel: int = 1,
        verbose: bool | str = False,
        **readargs,
    ) -> Self:
        """Find and read data from frames for a number of channels.

        This method uses :mod:`gwdatafind` to discover the (`file://`) URLs
        that provide the requested data, then reads those files using
        :meth:`TimeSeriesDict.read()`.

        This is just a shim around ``TimeSeriesDict.get(..., source="gwdatafind")``.

        Parameters
        ----------
        channels : `list`
            List of names of data channels to find.

        start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS start time of required data,
            any input parseable by `~gwpy.time.to_gps` is fine

        end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
            GPS end time of required data, defaults to end of data found;
            any input parseable by `~gwpy.time.to_gps` is fine

        observatory : `str`, optional
            The observatory to use when searching for data.
            Default is to use the observatory from the channel name prefix,
            but this should be specified when searching for data in a
            multi-observatory dataset (e.g. `observatory='HLV'`).

        frametype : `str`, optional
            Name of frametype (dataset) in which this channel is stored.
            Default is to search all available datasets for a match, which
            can be very slow.

        frametype_match : `str`, optional
            Regular expression to use for frametype matching.

        host : `str`, optional
            Name of the GWDataFind server to use.
            Default is set by `gwdatafind.utils.get_default_host`.

        urltype : `str`, optional
            The URL type to use.
            Default is "file" to use paths available on the file system.

        ext : `str`, optional
            The file extension for which to search.
            "gwf" is the only file extension supported, but this may be
            extended in the future.

        pad : `float`, optional
            Value with which to fill gaps in the source data,
            by default gaps will result in a `ValueError`.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        parallel : `int`, optional
            Number of parallel threads to use when reading data.

        allow_tape : `bool`, optional
            Allow reading from frame files on (slow) magnetic tape.

        verbose : `bool`, optional
            This argument is deprecated and will be removed in a future release.
            Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

        readargs
            Any other keyword arguments to be passed to `.read()`.

        Raises
        ------
        requests.exceptions.HTTPError
            If the GWDataFind query fails for any reason.

        RuntimeError
            If no files are found to read, or if the read operation
            fails.
        """
        return cls.get(
            channels,
            start,
            end,
            source="gwdatafind",
            observatory=observatory,
            frametype=frametype,
            frametype_match=frametype_match,
            host=host,
            urltype=urltype,
            ext=ext,
            pad=pad,
            scaled=scaled,
            allow_tape=allow_tape,
            parallel=parallel,
            verbose=verbose,
            **readargs,
        )

    @classmethod
    def from_arrakis(
        cls,
        block: arrakis.SeriesBlock,
        *,
        copy: bool = True,
        **metadata,
    ) -> Self:
        """Construct a new dict from an `arrakis.SeriesBlock`.

        Parameters
        ----------
        block : `arrakis.SeriesBlock`
            The input Arrakis data to read.

        copy : `bool`, optional
            If `True`, copy the contained data array to new  to a new array.

        metadata
            Any other metadata keyword arguments to pass to the `TimeSeries`
            constructor.

        Returns
        -------
        dict : `TimeSeriesDict`
            A new `TimeSeriesDict` containing the data from the Arrakis block.
        """
        tsd = cls()
        for name, series in block.items():
            tsd[name] = tsd.EntryClass.from_arrakis(
                series,
                copy=copy,
                **metadata,
            )
        return tsd

    @classmethod
    def from_nds2_buffers(
        cls,
        buffers: Iterable[nds2.buffer],
        *,
        scaled: bool | None = None,
        copy: bool = True,
        **metadata,
    ) -> Self:
        """Construct a new dict from a list of `nds2.buffer` objects.

        **Requires:** |nds2|_

        Parameters
        ----------
        buffers : `list` of `nds2.buffer`
            The input NDS2-client buffers to read.

        scaled : `bool`, optional
            Apply slope and bias calibration to ADC data, for non-ADC data
            this option has no effect.

        copy : `bool`, optional
            If `True`, copy the contained data array to new  to a new array.

        metadata
            Any other metadata keyword arguments to pass to the `TimeSeries`
            constructor.

        Returns
        -------
        dict : `TimeSeriesDict`
            A new `TimeSeriesDict` containing the data from the given buffers.
        """
        tsd = cls()
        for buf in buffers:
            tsd[buf.channel.name] = tsd.EntryClass.from_nds2_buffer(
                buf,
                scaled=scaled,
                copy=copy,
                **metadata,
            )
        return tsd

    def plot(
        self,
        label: str = "key",
        method: str = "plot",
        figsize: tuple[float, float] = (12, 4),
        xscale: str = "auto-gps",
        **kwargs,
    ) -> Plot:
        """Plot the data for this `TimeSeriesBaseDict`.

        Parameters
        ----------
        label : `str`, optional
            Labelling system to use, or fixed label for all elements
            Special values include

            - ``'key'``: use the key of the `TimeSeriesBaseDict`,
            - ``'name'``: use the :attr:`~TimeSeries.name` of each element

            If anything else, that fixed label will be used for all lines.

        method : `str`, optional
            The plotting method to use. This can be any method supported by the
            underlying plotting library (e.g., Matplotlib).

        figsize : `tuple[float, float]`, optional
            The size of the figure to create, in inches.

        xscale : `str`, optional
            The scale of the x-axis. This can be one of

            - ``'linear'``: linear scale
            - ``'log'``: logarithmic scale
            - ``'auto-gps'``: automatically determine scale based on GPS time

        kwargs
            All other keyword arguments are passed to the plotter as appropriate.
        """
        kwargs.update({
            "method": method,
            "label": label,
            "figsize": figsize,
            "xscale": xscale,
        })

        # make plot
        from ..plot import Plot

        if kwargs.get("separate", False):
            plot = Plot(*self.values(), **kwargs)
        else:
            plot = Plot(self.values(), **kwargs)

        # update labels
        artmap = {
            "plot": "lines",
            "scatter": "collections",
        }
        artists = [
            x for ax in plot.axes
            for x in getattr(ax, artmap.get(method, "lines"))
        ]
        for key, artist in zip(self, artists, strict=True):
            if label.lower() == "name":
                lab = self[key].name
            elif label.lower() == "key":
                lab = str(key)
            else:
                lab = label
            artist.set_label(lab)

        return plot

    def step(
        self,
        label: str = "key",
        where: Literal["pre", "post", "mid"] = "post",
        figsize: tuple[float, float] = (12, 4),
        xscale: str = "auto-gps",
        **kwargs,
    ) -> Plot:
        """Create a step plot of this dict.

        Parameters
        ----------
        label : `str`, optional
            Labelling system to use, or fixed label for all elements.
            Special values include

            ``'key'``
                Use the key of the `TimeSeriesBaseDict`

            ``'name'``
                Use the :attr:`~TimeSeries.name` of each element

            If anything else, that fixed label will be used for all lines.

        where : `str`, optional
            The location of the step change. This can be one of

            - ``'pre'``: the step change occurs before the x value
            - ``'post'``: the step change occurs after the x value
            - ``'mid'``: the step change occurs at the midpoint of the x value

        figsize : `tuple[float, float]`, optional
            The size of the figure to create, in inches.

        xscale : `str`, optional
            The scale of the x-axis. This can be one of

            - ``'linear'``: linear scale
            - ``'log'``: logarithmic scale
            - ``'auto-gps'``: automatically determine scale based on GPS time

        kwargs
            All other keyword arguments are passed to the plotter as appropriate.
        """
        kwargs.setdefault(
            "drawstyle",
            f"steps-{where}",
        )
        tmp = cast("Self", type(self)())
        for key, series in self.items():
            tmp[key] = series.append(series.value[-1:], inplace=False)

        return tmp.plot(
            label=label,
            figsize=figsize,
            xscale=xscale,
            **kwargs,
        )


# -- TimeSeriesBaseList --------------

# Type variable for generic list entries
_T = TypeVar("_T", bound=TimeSeriesBase)


class TimeSeriesBaseList(list[_T], Generic[_T]):
    """Fancy list representing a list of `TimeSeriesBase`.

    The `TimeSeriesBaseList` provides an easy way to collect and organise
    `TimeSeriesBase` for a single `Channel` over multiple segments.

    Parameters
    ----------
    items
        Any number of `TimeSeriesBase`.

    Returns
    -------
    list
        A new `TimeSeriesBaseList`.

    Raises
    ------
    TypeError
        if any elements are not `TimeSeriesBase`
    """

    EntryClass: ClassVar[type[TimeSeriesBase]] = TimeSeriesBase

    def __init__(self, *items: _T) -> None:
        """Initialise a new list."""
        if len(items) == 1 and isinstance(items[0], (list, tuple)):
            super().__init__(*items)
        else:
            super().__init__()
            for item in items:
                self.append(item)

    @property
    def segments(self) -> SegmentList:
        """The `span` of each series in this list."""
        return SegmentList([item.span for item in self])

    def append(self, item: _T) -> None:
        """Add a new element to the end of the list."""
        if not isinstance(item, self.EntryClass):
            msg = f"Cannot append type '{type(item).__name__}' to {type(self).__name__}"
            raise TypeError(msg)

        super().append(item)

    append.__doc__ = list.append.__doc__

    def extend(self, item: Iterable[_T]) -> None:
        """Add multiple elements to the end of the list."""
        item = TimeSeriesBaseList(*item)
        super().extend(item)

    extend.__doc__ = list.extend.__doc__

    def coalesce(self) -> Self:
        """Merge contiguous elements of this list into single objects.

        This method implicitly sorts and potentially shortens this list.
        """
        self.sort(key=lambda ts: ts.t0.value)
        i = j = 0
        N = len(self)
        while j < N:
            this = self[j]
            j += 1
            if j < N and this.is_contiguous(self[j]) == 1:
                while j < N and this.is_contiguous(self[j]):
                    try:
                        this = self[i] = this.append(self[j])
                    except ValueError as exc:
                        if "cannot resize this array" in str(exc):
                            this = this.copy()
                            this = self[i] = this.append(self[j])
                        else:
                            raise
                    j += 1
            else:
                self[i] = this
            i += 1
        del self[i:]
        return self

    def join(
        self,
        pad: float | None = None,
        gap: Literal["raise", "ignore", "pad"] | None = None,
    ) -> _T:
        """Concatenate all of the elements of this list into a single object.

        Parameters
        ----------
        pad : `float`, optional
            Value with which to fill gaps in the source data,
            by default gaps will result in a `ValueError`.

        gap : `str`, optional
            What to do if there are gaps in the data, one of

            ``'raise'``
                Raise a `ValueError`

            ``'ignore'``
                Remove gap and join data

            ``'pad'``
                Pad gap with zeros

            If `pad` is given and is not `None`, the default is ``'pad'``,
            otherwise ``'raise'``.

        Returns
        -------
        series : `gwpy.types.TimeSeriesBase` subclass
             A single series containing all data from each entry in this list.

        See Also
        --------
        TimeSeries.append
            For details on how the individual series are concatenated together.
        """
        if not self:
            return self.EntryClass(numpy.empty((0,) * self.EntryClass._ndim))  # noqa: SLF001
        self.sort(key=lambda t: t.x0.value)
        out = self[0].copy()
        for series in self[1:]:
            out.append(series, gap=gap, pad=pad)
        return out

    @overload
    def __getitem__(self, key: SupportsIndex) -> _T: ...
    @overload
    def __getitem__(self, key: slice) -> Self: ...

    def __getitem__(self, key: SupportsIndex | slice) -> Self | _T:
        """Get an item from this list.

        If the key is a slice, return a new `TimeSeriesBaseList` containing
        the sliced elements, otherwise return the element at that index.
        """
        if isinstance(key, slice):
            return type(self)(*super().__getitem__(key))
        return super().__getitem__(key)

    def copy(self) -> Self:
        """Return a copy of this list with each element copied to new memory."""
        out = type(self)()
        for series in self:
            out.append(series.copy())
        return out



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the `gwpy.time` command-line interface."""

from unittest import mock

import pytest
from dateutil import tz

from .. import __main__ as gwpy_time_cli


@pytest.mark.parametrize(("args", "result"), [
    (["Jan 1 2010"], 946339215),
    (["Jan", "1", "2010"], 946339215),
    (["946339215"], "2010-01-01 00:00:00.000000 UTC"),
    (["1161887657", "--local"], "2016-10-30 13:34:00.000000 CDT"),
])
@mock.patch("dateutil.tz.tzlocal", return_value=tz.gettz("America/Chicago"))
def test_main(mock_tzlocal, args, result, capsys):
    """Test the `gwpy.time` command-line interface."""
    gwpy_time_cli.main(args)
    out, err = capsys.readouterr()
    assert not err
    assert out.rstrip() == str(result)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Whitening a `TimeSeries`
########################

Most data recorded from a gravitational-wave interferometer carry information
across a wide band of frequencies, typically up to a few kiloHertz, but
often it is the case that the low-frequency amplitude dwarfs that of the
high-frequency content, making discerning high-frequency features difficult.
This is especially true of the LIGO differential arm strain measurement,
which encodes any gravitational wave signals that are present.

We employ a technique called 'whitening' to normalize the power at all
frequencies so that excess power at any frequency is more obvious.

We demonstrate below the LIGO-Livingston gravitational-wave strain
measurement signal around |GW200129|_, the loudest signal as yet
detected by LIGO.
"""

# %%
# Data access
# -----------
# First, we use the :doc:`gwosc <gwosc:index>` Python client to get the GPS
# time of the event:

from gwosc.datasets import event_gps
gps = event_gps("GW200129_065458")

# %%
# Then we can import the `TimeSeries` object and fetch the strain data using
# :meth:`TimeSeries.get` in a window around that event:

from gwpy.timeseries import TimeSeries
data = TimeSeries.get("L1", int(gps) - 64, int(gps) + 64)

# %%
# To demonstrate the relative power across the frequency band, we can
# quickly estimate an Amplitude Spectral Density (ASD) for these data:

asd = data.asd(fftlength=8)
plot = asd.plot(
    xlim=(8, 1000),
    ylabel="Strain ASD [$1/\\sqrt{Hz}$]",
)
plot.show()

# %%
# Whitening
# ---------
# The ASD clearly shows the dominance in amplitude of the lowest frequency
# components of the data, where the seismic noise around the observatory
# is most impactful.
# We can now :meth:`~TimeSeries.whiten` the data to to normalise the
# amplitudes across the frequency range:

white = data.whiten(fftlength=8)

# %%
# and can `~TimeSeries.plot` both the original and whitened data around the
# event time:

from gwpy.plot import Plot
plot = Plot(
    data,
    white,
    separate=True,
    sharex=True,
    epoch=gps,
    xlim=(gps - 1, gps + 1),
)
plot.axes[0].set_ylabel("Strain amplitude", fontsize=16)
plot.axes[1].set_ylabel("Whitened strain amplitude", fontsize=16)
plot.show()

# %%
# The top figure is dominated by the low-frequency noise, whereas the
# whitened data below highlights a few spikes in the data at higher
# frequencies.
#
# We can zoom in very close around the event time:

plot = white.crop(gps - .1, gps + .1).plot(
    ylabel="Whitened strain amplitude",
)
plot.axes[0].set_epoch(gps)
plot.show()

# %%
# Here, centred around time 0.03 is the clear signature of a binary black hole
# merger, |GW200129|_.
# This signal is completely hidden in the unfiltered data, but the simple act
# of whitening has exposed the loudest gravitational-wave event ever detected!



# Copyright (c) 2018-2020 Louisiana State University
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Alex Urban <alexander.urban@ligo.org>
.. currentmodule:: gwpy.timeseries

Inject a known signal into a `TimeSeries`
#########################################

It can often be useful to add some known signal to an inherently random
or noisy timeseries. For example, one might want to examine what
would happen if a binary black hole merger signal occured at or near
the time of a glitch. In LIGO data analysis, this procedure is referred
to as an *injection*.

In the example below, we will create a stream of random, white Gaussian
noise, then inject a simulation of GW150914 into it at a known time.
"""

# %%
# First, we prepare 32-seconds of Gaussian noise:

from numpy import random
from gwpy.timeseries import TimeSeries
rng = random.default_rng(0)
noise = TimeSeries(rng.normal(scale=2, size=32 * 16384), sample_rate=16384)

# %%
# Then we can download a simulation of the GW150914 signal from GWOSC:

url = "https://gwosc.org/s/events/GW150914/P150914/fig2-unfiltered-waveform-H.txt"
signal = TimeSeries.read(url, format="txt")
signal.t0 = 16

# %%
# Note, since this simulation cuts off before a certain time, it is
# important to taper its ends to zero to avoid ringing artifacts.
# We can accomplish this using the
# :meth:`~gwpy.timeseries.TimeSeries.taper` method.

signal = signal.taper()

# %%
# Since the time samples overlap, we can inject this into our noise data
# using :meth:`~gwpy.types.series.Series.inject`:

data = noise.inject(signal)

# %%
# Finally, we can visualize the full process in the time domain:

from gwpy.plot import Plot
plot = Plot(noise, signal, data, separate=True, sharex=True, sharey=True)
for ax, text in [
    (plot.axes[0], "Noise only"),
    (plot.axes[1], "Signal only"),
    (plot.axes[2], "Noise + signal"),
]:
    ax.text(
        0.01, .92, text,
        transform=ax.transAxes, ha="left", va="top",
        bbox={"facecolor": "white", "alpha": 0.8},
    )
ax1 = plot.axes[1]
ax1.set_ylabel("Amplitude")
ax1.set_epoch(0)
axins = ax1.inset_axes(
    [0.7, 0.4, 0.29, 0.55],
    xlim=(15.95, 16.25),
    ylim=(-2, 2),
    xticklabels=[],
    yticklabels=[],
)
axins.plot(signal)
ax1.indicate_inset_zoom(axins, edgecolor="black")
plot.show()

# %%
# Given the difference in amplitude, we can't see the injected signal in the
# noisy data at all.
# However, we can use the Q-transform to visualize things in the time-frequency
# domain:

outseg = (15.4, 16.4)
noiseq = noise.q_transform(outseg=outseg, fftlength=4)
dataq = data.q_transform(outseg=outseg, fftlength=4)
plot = Plot(
    noiseq,
    dataq,
    method="pcolormesh",
    separate=True,
    sharex=True,
    sharey=True,
    clim=(0, 25),
    yscale="log",
    ylim=(16, 1000),
)
for ax, text in [
    (plot.axes[0], "Noise only"),
    (plot.axes[1], "With injected signal"),
]:
    ax.text(
        0.01, .95, text,
        transform=ax.transAxes, ha="left", va="top",
        bbox={"facecolor": "white", "alpha": 0.8},
    )
plot.colorbar(label="Normalized Energy")
plot.show()

# %%
# Here, we can clearly see the injected signal in the Q-transform of the
# data.
#
# For more information on the Q-transform, see :doc:`qscan`.



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.omicron`."""

import pytest

from ...testing.utils import assert_table_equal
from .utils import (
    TABLE_CLASSES,
    random_table,
)

# -- test config ---------------------

def pytest_generate_tests(metafunc):
    """Parametrize the ``table`` fixture for each valid table type."""
    if "table" in metafunc.fixturenames:
        metafunc.parametrize("table", TABLE_CLASSES, indirect=True)


@pytest.fixture
def table(request):
    """Create a random table with the right columns."""
    return random_table(
        length=100,
        names=["time", "frequency", "snr"],
        table_class=request.param,
    )


@pytest.fixture
def root_file(table, tmp_path):
    """Write the ``table`` to a (temporary) file and return the path.

    Enables read tests.
    """
    tmp = tmp_path / "table.root"
    table.write(tmp, treename="triggers")
    return tmp


# -- tests ---------------------------

@pytest.mark.requires("uproot")
def test_read_write_root_omicron(table, root_file):
    """Test `Table.read(format='root.omicron')`."""
    # check read gives back same table
    t2 = type(table).read(root_file, format="root.omicron")
    assert_table_equal(t2, table)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Primitive command-line interface to `gwpy.time.tconvert`.

Either pass a GPS time to convert to a date string, or a date string
to convert to a GPS time.
"""

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from dateutil import tz

from .. import __version__
from ..tools import _utils
from . import tconvert

if TYPE_CHECKING:
    from argparse import ArgumentParser

EXAMPLES = {
    "Convert GPS time to date string": "gwpy-tconvert 1126259462",
    "Convert date string to GPS time": "gwpy-tconvert 2015-09-14 09:50:45",
    "Find GPS time now": "gwpy-tconvert now",
    "Find GPS time for the start of today": "gwpy-tconvert today",
}


def create_parser() -> ArgumentParser:
    """Parse command-line arguments, tconvert inputs, and print."""
    # define command line arguments
    parser = _utils.ArgumentParser(
        description=__doc__,
        examples=EXAMPLES,
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=__version__,
        help="show version number and exit",
    )
    parser.add_argument(
        "-l",
        "--local",
        action="store_true",
        default=False,
        help="print datetimes in local timezone",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        action="store",
        default=r"%Y-%m-%d %H:%M:%S.%f %Z",
        help="output datetime format (default: %(default)r)")
    parser.add_argument(
        "input",
        help="GPS or datetime string to convert",
        nargs="*",
    )
    return parser


def main(args: list[str] | None = None) -> None:
    """Run this tool."""
    # parse and convert
    parser = create_parser()
    opts = parser.parse_args(args)
    input_ = " ".join(opts.input)
    output = tconvert(input_)

    # print (now with timezones!)
    if isinstance(output, datetime.datetime):
        output = output.replace(tzinfo=tz.tzutc())
        if opts.local:
            output = output.astimezone(tz.tzlocal())
        print(output.strftime(opts.format))
    else:
        print(output)


if __name__ == "__main__":  # pragma: no-cover
    main()



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.rc`."""

from unittest import mock

import pytest
from matplotlib import rcParams

from .. import rc as plot_rc

DEFAULT_LRTB = [
    rcParams[f"figure.subplot.{x}"]
    for x in ("left", "right", "bottom", "top")
]


@pytest.mark.parametrize("usetex", [
    False,
    True,
    None,
])
@mock.patch("gwpy.plot.rc.bool_env", return_value=False)
def test_rc_params(mock_bool_env, usetex):
    """Test `rc_params()`."""
    rcp = plot_rc.rc_params(usetex=usetex)
    assert rcp["axes.edgecolor"] == plot_rc.GWPY_RCPARAMS["axes.edgecolor"]
    assert rcp.get("text.usetex", False) == bool(usetex)


@pytest.mark.parametrize(("figsize", "lrbt"), [
    pytest.param(
        (6.4, 4.8),
        (.1875, .87, .16, .88),
        id="standard",
    ),
    pytest.param(
        (0, 0),
        DEFAULT_LRTB,
        id="default",
    ),
])
def test_get_subplot_params(figsize, lrbt):
    """Test `get_subplot_params()`."""
    params = plot_rc.get_subplot_params(figsize)
    for key, val in zip(("left", "right", "bottom", "top"), lrbt, strict=True):
        assert getattr(params, key) == val



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Calculating the time-dependent coherence between two channels
#############################################################

The standard coherence calculation outputs a frequency series
(`~gwpy.frequencyseries.FrequencySeries`) giving a time-averaged measure
of coherence. See :ref:`sphx_glr_examples_frequencyseries_coherence.py` for an
example.

The `TimeSeries` method :meth:`~TimeSeries.coherence_spectrogram` performs the
same coherence calculation every ``stride``, giving a time-varying coherence
measure.

These data are available as part of the |GWOSC_O3_AUX_RELEASE|_.
"""

# %%
# First, we import the `TimeSeries`
from gwpy.timeseries import TimeSeries

# %%
# and then :meth:`~TimeSeries.get` the LIGO-Hanford strain data,
# and the mains power monitor data, for a 20 minute (1200 second
# window near the end of the |GWOSC_O3_AUX_RELEASE|_:

start = 1389410000
end = 1389411200
strain = TimeSeries.get("H1", start, end)
mains = TimeSeries.get(
    "H1:PEM-EY_MAINSMON_EBAY_1_DQ",
    start,
    end,
    frametype="H1_AUX_AR1",
)

# %%
#
# .. admonition:: Data are accessed separately
#     :class: note
#
#     We could have used `TimeSeriesDict.get` to access both channels in a
#     single call - that method would first try to find a single dataset that
#     contain both of them, then automatically fall back to separate calls if
#     that fails.
#
#     But, since we know that these channels are in different datsets, we
#     access them separately here for clarity and speed.
#
# Calculating coherence
# ---------------------
# We can then calculate the :meth:`~TimeSeries.coherence` of one
# `TimeSeries` with respect to the other, using an 2-second Fourier
# transform length, with a 1-second (50%) overlap, and a stride of 10:
coh = strain.coherence_spectrogram(mains, 20, fftlength=2, overlap=1)

# %%
# Finally, we can :meth:`~gwpy.spectrogram.Spectrogram.plot` the
# resulting data
plot = coh.plot()
ax = plot.gca()
ax.set_ylabel("Frequency [Hz]")
ax.set_yscale("log")
ax.set_ylim(10, 2000)
ax.set_title("Coherence between Power mains and LIGO-Hanford strain data")
ax.grid(visible=True, which="both", axis="both")
ax.colorbar(label="Coherence", clim=[0, 1], cmap="plasma")
plot.show()

# %%
# This shows the time-dependent coherence between the two channels.
# The 60 Hz power mains frequency and its harmonics are clearly visible,
# but the coherence is variable over time, demonstrating the varying
# fidelity of the incoming mains signal.



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""LIGOTimeGPS object discovery."""

# ruff: noqa: I001

from __future__ import annotations

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

from typing import (
    TYPE_CHECKING,
    Protocol,
    runtime_checkable,
)

if TYPE_CHECKING:
    from typing import Self

# try and import LIGOTimeGPS from LAL, otherwise use the pure-python backup
# provided by the ligotimegps package, its slower, but works
try:
    # import h5py first to avoid https://git.ligo.org/lscsoft/lalsuite/-/issues/821
    import h5py  # noqa: F401
    # then import LAL
    from lal import LIGOTimeGPS
except ImportError:
    from ligotimegps import LIGOTimeGPS

__all__ = [
    "LIGOTimeGPS",
    "LIGOTimeGPSLike",
]


@runtime_checkable
class LIGOTimeGPSLike(Protocol):
    """Protocol for types that are implementations of LIGOTimeGPS.

    This is used for type hinting functions that can accept
    |lal.LIGOTimeGPS|_, or `ligotimegps.LIGOTimeGPS`, or any
    other implementation of the ``LIGOTimeGPS`` standard.

    This can also be used at runtime with `isinstance` to check if an
    object is a LIGOTimeGPS-like object, but cannot be used with
    `issubclass`.
    """

    gpsSeconds: int      # noqa: N815
    gpsNanoSeconds: int  # noqa: N815

    def __init__(self, new: float | LIGOTimeGPS | str) -> None: ...

    def __float__(self) -> float: ...

    def __add__(self, other: Self | float) -> Self: ...
    def __radd__(self, other: Self | float) -> Self: ...



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for generating colour bars for figures.

This module mainly exists to support generating colour bars for figures
where the 'parent' Axes isn't resized to accommodate the new Axes for
the colour bar.

GWpy adds this functionality by overloading the `Figure.colorbar` method
with calls to functions below that add support for handling ``fraction=0.``.
"""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING

from matplotlib.colors import LogNorm
from matplotlib.figure import Figure
from matplotlib.ticker import LogFormatterSciNotation

from .colors import format_norm
from .log import LogFormatter

if TYPE_CHECKING:
    from typing import (
        Any,
        Literal,
    )

    from matplotlib.axes import Axes
    from matplotlib.collections import Collection
    from matplotlib.colorbar import Colorbar
    from matplotlib.image import AxesImage

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- custom colorbar generation ------

def _process_colorbar_kwargs(
    figure: Figure,
    mappable: AxesImage | Collection | None = None,
    ax: Axes | None = None,
    cax: Axes | None = None,
    fraction: float = 0.,
    **kwargs,
) -> tuple[AxesImage | Collection, dict[str, Any]]:
    """Configure the keyword arguments for colorbars.

    The main purpose of this function is to replace the default matplotlib
    behaviour (resizing the 'parent' axes to make space for the colorbar
    axes) with our default or creating a new axes alongside the parent
    axes without resizing.

    Parameters
    ----------
    figure : `matplotlib.figure.Figure`
        The figure on which to draw the new colorbar axes.

    mappable : `~matplotlib.image.AxesImage`, `~matplotlib.collections.Collection`
        Collection against which to map the colouring, default will
        be the last added mappable artist (collection or image)

    ax : `matplotlib.axes.Axes`
        The `Axes` against which to anchor the colorbar Axes.

    cax : `matplotlib.axes.Axes`
        The `Axes` on which to draw the colorbar.

    fraction : `float`
        The fraction of space to steal from the parent Axes to make
        space for the colourbar.

    kwargs
        Other keyword arguments are passed to
        :meth:`matplotlib.figure.Figure.colorbar`

    Returns
    -------
    mappable : `~matplotlib.image.AxesImage`, `~matplotlib.collections.Collection`
        The collection against which to map the colouring.

    kwargs
        A dict of keyword arguments to pass to
        :meth:`matplotlib.figure.Figure.colorbar`.
    """
    # get mappable and axes objects
    if mappable is None and ax is None:
        mappable = find_mappable(*figure.axes)
    if mappable is None:
        mappable = find_mappable(ax)
    if ax is None:
        ax = mappable.axes

    # -- format mappable

    # parse normalisation
    norm, kwargs = format_norm(kwargs, mappable.norm)
    mappable.set_norm(norm)
    mappable.set_cmap(kwargs.pop("cmap", mappable.get_cmap()))

    # -- set tick formatting

    if isinstance(norm, LogNorm):
        kwargs.setdefault("format", LogFormatter())

    # -- create axes for colorbar (if required)

    if cax is not None:  # cax was given, we don't need fraction
        kwargs.pop("fraction", None)
    elif fraction == 0.:  # if fraction is 0, make the inset axes ourselves
        cax, kwargs = _make_inset_axes(ax, **kwargs)
    else:  # otherwise let matplotlib generate the Axes using its own default
        kwargs["fraction"] = fraction

    # pack kwargs
    kwargs.update(ax=ax, cax=cax)
    return mappable, kwargs


def colorbar(
    figure: Figure,
    ax: Axes | None = None,
    mappable: AxesImage | Collection | None = None,
    cax: Axes | None = None,
    fraction: float = 0.,
    *,
    emit: bool = True,
    **kwargs,
) -> Colorbar:
    """Add a colorbar to the current ``figure``.

    This method differs from the default
    :meth:`matplotlib.figure.Figure.colorbar` in that it doesn't
    resize the parent `Axes` to accommodate the colorbar, but rather
    draws a new Axes alongside it.

    Parameters
    ----------
    figure : `~matplotlib.figure.Figure`
        The figure onto which to draw the colorbar.

    ax : `~matplotlib.axes.Axes`, optional
        Axes relative to which to position colorbar.
        The default is the `Axes` containing the ``mappable``.

    mappable : `~matplotlib.image.AxesImage`, `~matplotlib.collections.Collection`, optional
        Collection against which to map the colorbar.
        Default is the most-recently-added mappable artist.

    cax : `~matplotlib.axes.Axes`, optional
        Axes on which to draw colorbar.
        By default a new `Axes` will be created.

    fraction : `float`, optional
        Fraction of original axes to use for colorbar.
        The default (``fraction=0``) is to not resize the
        original axes at all.

    emit : `bool`, optional
        If `True` update all mappables on `Axes` to match the same
        colouring as the colorbar.

    kwargs
        Other keyword arguments are passed to
        :meth:`~matplotlib.figure.Figure.colorbar`.

    Returns
    -------
    cbar : `~matplotlib.colorbar.Colorbar`
        The newly added `Colorbar`.

    Notes
    -----
    To revert to the default matplotlib behaviour, pass ``fraction=0.15``.

    See Also
    --------
    matplotlib.figure.Figure.colorbar
    matplotlib.colorbar.Colorbar

    Examples
    --------
    >>> import numpy
    >>> from gwpy.plot import Plot

    To plot a simple image and add a colorbar:

    >>> plot = Plot()
    >>> ax = plot.gca()
    >>> ax.imshow(numpy.random.randn(120).reshape((10, 12)))
    >>> plot.colorbar(label='Value')
    >>> plot.show()

    Colorbars can also be generated by directly referencing the parent
    axes:

    >>> Plot = Plot()
    >>> ax = plot.gca()
    >>> ax.imshow(numpy.random.randn(120).reshape((10, 12)))
    >>> ax.colorbar(label='Value')
    >>> plot.show()
    """
    # pre-process kwargs (and maybe create new Axes)
    mappable, kwargs = _process_colorbar_kwargs(
        figure,
        ax=ax,
        cax=cax,
        fraction=fraction,
        mappable=mappable,
        **kwargs,
    )

    # generate colour bar using upstream method
    cbar = Figure.colorbar(figure, mappable, **kwargs)

    # force the minor ticks to be the same as the major ticks;
    # in practice, this normally swaps out LogFormatterSciNotation to
    # gwpy's LogFormatter;
    # this is hacky, and would be improved using a
    # subclass of Colorbar in the first place, but matplotlib's
    # cbar_factory doesn't support that
    longaxis = (
        cbar.ax.yaxis if cbar.orientation == "vertical"
        else cbar.ax.xaxis
    )
    if (
        isinstance(cbar.formatter, LogFormatter)
        and isinstance(
            longaxis.get_minor_formatter(),
            LogFormatterSciNotation,
        )
    ):
        longaxis.set_minor_formatter(type(cbar.formatter)())

    # update mappables for this axis
    if emit:
        ax: Axes = kwargs["ax"]
        norm = mappable.norm
        cmap = mappable.get_cmap()
        for map_ in ax.collections + ax.images:
            map_.set_norm(norm)
            map_.set_cmap(cmap)

    return cbar


# -- utilities -----------------------

def find_mappable(*axes: Axes) -> Collection | AxesImage:
    """Find the most recently added mappable layer in the given axes.

    Parameters
    ----------
    axes : `~matplotlib.axes.Axes`
        One or more axes to search for a mappable.

    Returns
    -------
    mappable :

    Raises
    ------
    ValueError
        If a collection or an image cannot be found to map a colorbar onto.
    """
    for ax in axes:
        for aset in ("collections", "images"):
            with contextlib.suppress(AttributeError, IndexError):
                return getattr(ax, aset)[-1]
    msg = "cannot determine mappable layer on any axes for this colorbar"
    raise ValueError(msg)


def _scale_width(value: float, ax: Axes) -> float:
    """Scale a width value based on the width of the axes."""
    fig = ax.figure
    return value / (ax.get_position().width * fig.get_figwidth())


def _scale_height(value: float, ax: Axes) -> float:
    """Scale a height value based on the height of the axes."""
    fig = ax.figure
    return value / (ax.get_position().height * fig.get_figheight())


def _colorbar_bounds(
    ax: Axes,
    location: Literal["left", "right", "top", "bottom"] = "right",
    width: float | None = None,
    length: float = 1.,
    orientation: Literal["horizontal", "vertical"] | None = None,
    pad: float | None = None,
) -> tuple[float, float, float, float]:
    """Return the ``bounds`` for an inset Axes designed for a colourbar.

    Parameters
    ----------
    ax : `matplotlib.axes.Axes`
        The axes to anchor to.

    location : `str`
        Where to place the colourbar, one of

        - ``"left"``
        - ``"right"`` (default)
        - ``"top"``
        - ``"bottom"``

    width : `float`
        The size of the colourbar along its short axis (i.e. 'width' for
        a vertical bar, 'height' for a horizontal bar), as a fraction of
        the parent `Axes` size in the same direction.

    length : `float`
        The size of the colourbar along its long axis (i.e. 'height' for
        a vertical bar, 'length' for a horizontal bar), as a fraction of
        the parent `Axes` size in the same direction.

    orientation : `str` or `None`
        One of ``"horizontal"`` or ``"vertical"``. The default is
        ``"horizontal"`` if ``location`` is ``"top"`` or ``"bottom"``,
        otherwise ``"vertical"``.

    pad : `float`
        The gap between the parent Axes and the colourbar, as a fraction
        of the parent Axes length in the relevant direction. Default is
        equivalent to .1 inches.

    Returns
    -------
    bounds : `tuple` of `float`
        The ``(x0, y0, width, height)`` bounds for an inset Axes to
        contain a colourbar. This is designed to be passed to
        :meth:`~matplotlib.axes.Axes.inset_axes`.
    """
    location = location.lower()

    # calculate default width and padding for the relevant orientation
    orientation = "vertical" if location in {"left", "right"} else "horizontal"
    if orientation == "vertical":
        size = _scale_width(.1, ax)
    elif orientation == "horizontal":
        size = _scale_height(.1, ax)
    pad = size if pad is None else pad
    width = size if width is None else width

    # where to start on the long axis (for centre-aligned Axes)
    l0 = .5 - length / 2.

    if location == "left":
        return 0 - pad - width, l0, width, length
    if location == "right":
        return 1 + pad, l0, width, length
    if location == "top":
        return l0, 1 + pad, length, width
    # "bottom":
    return l0, 0 - pad - width, length, width


def _make_inset_axes(
    ax: Axes,
    location: Literal["left", "right", "top", "bottom"] = "right",
    width: float | None = None,
    length: float = 1.,
    pad: float | None = None,
    **kwargs,
) -> tuple[Axes, dict[str, Any]]:
    """Create a new `Axes` to support a colorbar using `Axes.inset_axes`."""
    # set default orientation
    if location in {"left", "right"}:
        orientation = kwargs.setdefault("orientation", "vertical")
    elif location in {"top", "bottom"}:
        orientation = kwargs.setdefault("orientation", "horizontal")
    else:
        msg = f"inset Axes location '{location}' not recognised"
        raise ValueError(msg)

    # set default ticklocation for the right location/orientation
    if location == "top" and orientation == "horizontal":
        kwargs.setdefault("ticklocation", "top")
    if location == "left" and orientation == "vertical":
        kwargs.setdefault("ticklocation", "left")

    # calculate default width as a fixed size
    if width is None:
        width = _scale_width(0.1, ax)

    # calculate the bounds for the new Axes
    bounds = _colorbar_bounds(
        ax,
        location=location,
        width=width,
        length=length,
        orientation=orientation,
        pad=pad,
    )

    return ax.inset_axes(bounds, transform=ax.transAxes), kwargs



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.omega`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from ...testing.utils import assert_table_equal
from ..table import EventTable

if TYPE_CHECKING:
    from pathlib import Path

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


OMEGA_ASCII = """
% time
% frequency
% duration
% bandwidth
% normalizedEnergy
1023927596.843750000 1.1766601562500000e+01 1.8193577533403701e-01 5.4964450953309427e+00 2.7199652890340332e+01
1023927378.250000000 2.3637695312500000e+00 9.0565757364173183e-01 1.1041700849239395e+00 2.1194302085268539e+01
1023927597.187500000 1.0293457031250000e+01 2.0797345069017714e-01 4.8083060442639054e+00 1.7678989962812786e+01
1023927433.250000000 2.3637695312500000e+00 9.0565757364173183e-01 1.1041700849239395e+00 1.6883206146877814e+01
""".strip()  # noqa: E501


@pytest.fixture
def omega_ascii(tmp_path: Path) -> Path:
    """Create a temporary file with example Omega ascii data."""
    dat = tmp_path / "EVENTS.txt"
    dat.write_text(OMEGA_ASCII)
    return dat


def test_read_omega_ascii(omega_ascii):
    """Check that the ``"ascii.omega"`` file reader works."""
    tab = EventTable.read(omega_ascii, format="ascii.omega")
    assert len(tab) == 4
    assert tab[0]["time"] == pytest.approx(1023927596.84375)
    assert tab[3]["normalizedEnergy"] == pytest.approx(16.8832)


def test_read_omega_ascii_noheader(tmp_path):
    """Check that reading an Omega file without headers results in an error."""
    # create empty file
    empty = tmp_path / "empty.txt"
    empty.write_text("")

    # read it back
    with pytest.raises(
        ValueError,
        match="No column names found in Omega header",
    ):
        EventTable.read(empty, format="ascii.omega")


def test_write_omega_ascii(omega_ascii):
    """Check that the ``"ascii.omega"`` file writer works."""
    # read the example data
    tab = EventTable.read(omega_ascii, format="ascii.omega")

    # write the table out in ascii.omega format
    out = omega_ascii.parent / "write.txt"
    tab.write(out, format="ascii.omega")

    # then read it from the written file and compare
    t2 = EventTable.read(out, format="ascii.omega")

    assert_table_equal(t2, tab)



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for progress bars."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from tqdm import tqdm

if TYPE_CHECKING:
    from typing import TextIO

TQDM_BAR_FORMAT: str = (
    "{desc}: |{bar}| "
    "{n_fmt}/{total_fmt} ({percentage:3.0f}%) "
    "ETA {remaining:6s}"
)


def progress_bar(
    desc: str = "Processing",
    file: TextIO = sys.stdout,
    bar_format: str = TQDM_BAR_FORMAT,
    **kwargs,
) -> tqdm:
    """Create a `tqdm.tqdm` progress bar.

    This is just a thin wrapper around `tqdm.tqdm` with some updated defaults.
    """
    pbar = tqdm(
        desc=desc,
        file=file,
        bar_format=bar_format,
        **kwargs,
    )
    if not pbar.disable:
        pbar.desc = pbar.desc.rstrip(": ")
        pbar.refresh()
    return pbar



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Representation of a frequency series."""

from __future__ import annotations

import warnings
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy
from astropy import units
from numpy import fft as npfft

from ..io.registry import UnifiedReadWriteMethod
from ..types import Series
from ..utils.misc import property_alias
from .connect import (
    FrequencySeriesRead,
    FrequencySeriesWrite,
)

if TYPE_CHECKING:
    from typing import (
        ClassVar,
        Self,
    )

    import pycbc.types
    from astropy.units import (
        Quantity,
        UnitBase,
    )

    from ..detector import Channel
    from ..plot import Plot
    from ..signal.filter_design import FilterCompatible
    from ..time import SupportsToGps
    from ..timeseries import TimeSeries
    from ..typing import (
        ArrayLike1D,
        QuantityLike,
        UnitLike,
    )
    from ..utils.lal import LALFrequencySeriesType

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class FrequencySeries(Series):
    """A data array holding some metadata to represent a frequency series.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`, optional
        Physical unit of these data.

    f0 : `float`, `~astropy.units.Quantity`, optional, default: `0`
        Starting frequency for these data.

    df : `float`, `~astropy.units.Quantity`, optional, default: `1`
        Frequency resolution for these data.

    frequencies : `array-like`
        The complete array of frequencies indexing the data.
        This argument takes precedence over `f0` and `df` so should
        be given in place of these if relevant, not alongside.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional, default: `False`
        Choose to copy the input data to new memory.

    subok : `bool`, optional, default: `True`
        Allow passing of sub-classes by the array generator.

    Notes
    -----
    Key methods:

    .. autosummary::

       ~FrequencySeries.read
       ~FrequencySeries.write
       ~FrequencySeries.plot
       ~FrequencySeries.zpk
    """

    _default_xunit: ClassVar[UnitBase] = units.hertz
    _print_slots: ClassVar[tuple[str, ...]] = (
        "f0",
        "df",
        "epoch",
        "name",
        "channel",
    )

    def __new__(
        cls,
        data: ArrayLike1D,
        unit: UnitLike = None,
        f0: Quantity | float | None = None,
        df: Quantity | float | None = None,
        frequencies: ArrayLike1D | None = None,
        name: str | None = None,
        epoch: SupportsToGps | None = None,
        channel: Channel | str | None = None,
        **kwargs,
    ) -> Self:
        """Generate a new FrequencySeries."""
        if f0 is not None:
            kwargs["x0"] = f0
        if df is not None:
            kwargs["dx"] = df
        if frequencies is not None:
            kwargs["xindex"] = frequencies

        # generate FrequencySeries
        return super().__new__(
            cls,
            data,
            unit=unit,
            name=name,
            channel=channel,
            epoch=epoch,
            **kwargs,
        )

    # -- FrequencySeries properties --

    f0 = property_alias(Series.x0, "Starting frequency for this `FrequencySeries`")  # type: ignore[arg-type]
    df = property_alias(Series.dx, "Frequency spacing of this `FrequencySeries`")  # type: ignore[arg-type]
    frequencies = property_alias(Series.xindex, "Series of frequencies for each sample")  # type: ignore[arg-type]

    # -- FrequencySeries i/o ---------

    read = UnifiedReadWriteMethod(FrequencySeriesRead)
    write = UnifiedReadWriteMethod(FrequencySeriesWrite)

    # -- FrequencySeries methods -----

    def plot(
        self,
        method: str = "plot",
        xscale: str = "log",
        **kwargs,
    ) -> Plot:
        """Plot the data for this `FrequencySeries`."""
        # use log y-scale for ASD, PSD
        u = self.unit
        try:
            hzpow = u.powers[u.bases.index(units.Hz)]  # type: ignore[union-attr]
        except ValueError:
            pass
        else:
            if hzpow < 0:
                kwargs.setdefault("yscale", "log")

        return super().plot(method=method, xscale=xscale, **kwargs)

    def ifft(self) -> TimeSeries:
        """Compute the one-dimensional discrete inverse Fourier transform.

        Returns
        -------
        out : :class:`~gwpy.timeseries.TimeSeries`
            The normalised, real-valued `TimeSeries`.

        See Also
        --------
        numpy.fft.irfft
            The inverse (real) FFT function.

        Notes
        -----
        This method applies the necessary normalisation such that the
        condition holds:

        >>> timeseries = TimeSeries([1.0, 0.0, -1.0, 0.0], sample_rate=1.0)
        >>> timeseries.fft().ifft() == timeseries
        """
        from ..timeseries import TimeSeries

        nout = (self.size - 1) * 2
        # Undo normalization from TimeSeries.fft
        # The DC component does not have the factor of two applied
        # so we account for it here
        array = self.value.copy()
        array[1:] /= 2.0
        dift = npfft.irfft(
            array * nout,
            n=nout,
        )
        return TimeSeries(
            dift,
            epoch=self.epoch,
            channel=self.channel,
            name=self.name,
            unit=self.unit,
            dx=(1 / self.dx / nout).to("s"),
        )

    def zpk(
        self,
        zeros: ArrayLike1D,
        poles: ArrayLike1D,
        gain: float,
        *,
        analog: bool = True,
        sample_rate: QuantityLike | None = None,
        unit: str = "rad/s",
        normalize_gain: bool = False,
    ) -> Self:
        """Filter this `FrequencySeries` by applying a zero-pole-gain filter.

        Parameters
        ----------
        zeros : `array-like`
            List of zero frequencies (in Hertz).

        poles : `array-like`
            List of pole frequencies (in Hertz).

        gain : `float`
            DC gain of filter.

        analog : `bool`, optional
            Type of ZPK being applied, if ``analog=True`` all parameters
            will be converted in the Z-domain for digital filtering via
            the bilinear transform.

        sample_rate : `float`, `~astropy.units.Quantity`, optional
            Sample rate of data (in Hertz), used to apply a digital filter.
            Defaults to the last frequency value of this `FrequencySeries`
            (i.e. the Nyquist frequency).

        unit : `str`, optional
            For analogue ZPK filters, the units in which the zeros and poles are
            specified. Either ``'Hz'`` or ``'rad/s'`` (default).

        normalize_gain : `bool`, optional
            Whether to normalize the gain when converting from Hz to rad/s.

            - `False` (default):
              Multiply zeros/poles by -2π but leave gain unchanged.
              This matches the LIGO GDS **'f' plane** convention
              (``plane='f'`` in ``s2z()``).

            - `True`:
              Normalize gain to preserve frequency response magnitude.
              Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
              Use this when your filter was designed with the transfer
              function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
              This matches the LIGO GDS **'n' plane** convention
              (``plane='n'`` in ``s2z()``).

            Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

        Returns
        -------
        spectrum : `FrequencySeries`
            The frequency-domain filtered version of the input data.

        See Also
        --------
        FrequencySeries.filter
            For details on how a digital ZPK-format filter is applied.

        Examples
        --------
        To apply a zpk filter with file poles at 100 Hz, and five zeros at
        1 Hz (giving an overall DC gain of 1e-10)::

            >>> data2 = data.zpk([100]*5, [1]*5, 1e-10)
        """
        return self.filter(
            (zeros, poles, gain),
            sample_rate=sample_rate,
            analog=analog,
            unit=unit,
            normalize_gain=normalize_gain,
        )

    def interpolate(self, df: float) -> Self:
        """Interpolate this `FrequencySeries` to a new resolution.

        Parameters
        ----------
        df : `float`
            Desired frequency resolution of the interpolated `FrequencySeries`, in Hz.

        Returns
        -------
        out : `FrequencySeries`
            The interpolated version of the input `FrequencySeries`.

        See Also
        --------
        numpy.interp
            For the underlying 1-D linear interpolation scheme.
        """
        f0 = self.f0.decompose().value
        n_samples = (self.size - 1) * (self.df.decompose().value / df) + 1
        fsamples = numpy.arange(
            0,
            numpy.rint(n_samples),
            dtype=self.real.dtype,
        ) * df + f0
        out = type(self)(
            numpy.interp(fsamples, self.frequencies.value, self.value),
        )
        out.__array_finalize__(self)
        out.f0 = f0
        out.df = df
        return out

    def filter(
        self,
        filt: FilterCompatible,
        *,
        analog: bool = False,
        sample_rate: QuantityLike | None = None,
        unit: str = "rad/s",
        normalize_gain: bool = False,
        inplace: bool = False,
    ) -> Self:
        """Apply a filter to this `FrequencySeries`.

        The input filter argument is designed to accept any filter created
        by the :mod:`scipy.signal` filter design functions, and operates on
        the conventions of that module.

        Parameters
        ----------
        filt : `numpy.ndarray` or `tuple`
            The filter to be applied.
            This can be specified in any of the following forms, with
            the appropriate number of elements in the tuple:

            - `numpy.ndarray` - 1D array of FIR filter coefficients.
            - `tuple[numpy.ndarray, numpy.ndarray]` - numerator/demoinator
              polynomials of the transfer function.
            - `numpy.ndarray` - 2D array of SOS coefficients.
            - `tuple[numpy.ndarray, numpy.ndarray, float]` - zero-pole-gain
              representation.

        analog : `bool`, optional
            Type of ZPK being applied, if ``analog=True`` all parameters
            will be converted in the Z-domain for digital filtering via
            the bilinear transform.

        sample_rate : `float`, `~astropy.units.Quantity`, optional
            Sample rate of data (in Hertz), used to apply a digital filter.
            Defaults to the last frequency value of this `FrequencySeries`
            (i.e. the Nyquist frequency).

        unit : `str`, optional
            For analogue ZPK filters, the units in which the zeros and poles are
            specified. Either ``'Hz'`` or ``'rad/s'`` (default).

        normalize_gain : `bool`, optional
            Whether to normalize the gain when converting from Hz to rad/s.

            - `False` (default):
              Multiply zeros/poles by -2π but leave gain unchanged.
              This matches the LIGO GDS **'f' plane** convention
              (``plane='f'`` in ``s2z()``).

            - `True`:
              Normalize gain to preserve frequency response magnitude.
              Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
              Use this when your filter was designed with the transfer
              function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
              This matches the LIGO GDS **'n' plane** convention
              (``plane='n'`` in ``s2z()``).

            Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

        inplace : bool, optional
            If `True`, this array will be overwritten with the filtered
            version.

        Returns
        -------
        result : `FrequencySeries`
            The filtered version of the input `FrequencySeries`,
            if ``inplace=True`` was given, this is just a reference to
            the modified input array.

        Raises
        ------
        ValueError
            If ``filt`` arguments cannot be interpreted properly.

        See Also
        --------
        FrequencySeries.zpk
            For applying a zero-pole-gain filter, including in other
            units (e.g. poles and zeros specified in Hertz).
        """
        from ._fdcommon import _fdfilter
        return cast("Self", _fdfilter(
            self,
            filt,
            analog=analog,
            inplace=inplace,
            sample_rate=sample_rate,
            unit=unit,
            normalize_gain=normalize_gain,
        ))

    @classmethod
    def from_lal(
        cls,
        lalfs: LALFrequencySeriesType,
        *,
        copy: bool = True,
    ) -> Self:
        """Generate a new `FrequencySeries` from a LAL `FrequencySeries`.

        Any type of LAL FrequencySeries is supported.
        """
        # convert units
        from ..utils.lal import from_lal_unit

        try:
            unit = from_lal_unit(lalfs.sampleUnits)
        except TypeError:
            unit = None

        # create a new series
        return cls(
            lalfs.data.data,
            name=lalfs.name or None,
            unit=unit,
            f0=lalfs.f0,
            df=lalfs.deltaF,
            epoch=lalfs.epoch,
            channel=None,
            copy=copy,
        )

    def to_lal(self) -> LALFrequencySeriesType:
        """Convert this `FrequencySeries` into a LAL FrequencySeries.

        Returns
        -------
        lalspec : `FrequencySeries`
            An XLAL-format FrequencySeries of a given type, e.g.
            `REAL8FrequencySeries`.
        """
        import lal

        from ..utils.lal import (
            find_typed_function,
            to_lal_unit,
        )

        # map unit
        try:
            unit, scale = to_lal_unit(self.unit)
        except ValueError as exc:
            warnings.warn(
                f"{exc}, defaulting to lal.DimensionlessUnit",
                stacklevel=2,
            )
            unit = lal.DimensionlessUnit
            scale = 1

        # convert epoch
        epoch = lal.LIGOTimeGPS(0 if self.epoch is None else self.epoch.gps)

        # create FrequencySeries
        create = find_typed_function(self.dtype, "Create", "FrequencySeries")
        lalfs = create(
            self.name or "",
            epoch,
            self.f0.value,
            self.df.value,
            unit,
            self.shape[0],
        )
        lalfs.data.data = self.value * scale

        return lalfs

    @classmethod
    def from_pycbc(
        cls,
        fs: pycbc.types.TimeSeries,
        *,
        copy: bool = True,
    ) -> Self:
        """Convert a `pycbc.types.frequencyseries.FrequencySeries`.

        Parameters
        ----------
        fs : `pycbc.types.frequencyseries.FrequencySeries`
            The input PyCBC `~pycbc.types.frequencyseries.FrequencySeries` array.

        copy : `bool`, optional
            If `True`, copy these data to a new array.

        Returns
        -------
        spectrum : `FrequencySeries`
            A GWpy version of the input frequency series.
        """
        return cls(
            fs.data,
            f0=0,
            df=fs.delta_f,
            epoch=fs.epoch,
            copy=copy,
        )

    def to_pycbc(self, *, copy: bool = True) -> pycbc.types.FrequencySeries:
        """Convert this `FrequencySeries` into a PyCBC FrequencySeries.

        Parameters
        ----------
        copy : `bool`, optional
            If `True`, copy these data to a new array.

        Returns
        -------
        frequencyseries : `pycbc.types.frequencyseries.FrequencySeries`
            A PyCBC representation of this `FrequencySeries`.
        """
        from pycbc import types

        epoch = None if self.epoch is None else self.epoch.gps
        if self.f0.to("Hz").value:
            msg = (
                f"Cannot convert FrequencySeries to PyCBC with f0 = {self.f0}."
                " Starting frequency must be equal to 0 Hz."
            )
            raise ValueError(msg)
        return types.FrequencySeries(
            self.value,
            delta_f=self.df.to("Hz").value,
            epoch=epoch,
            copy=copy,
        )



# Copyright (c) 2021-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.timeseries.io.losc`."""

import pytest
from astropy.utils.data import download_file
from gwosc.locate import get_event_urls

from ...testing.errors import (
    pytest_skip_flaky_network,
    pytest_skip_network_error,
)
from ...utils.env import bool_env
from .. import (
    StateVector,
    TimeSeries,
)

GWPY_CACHE = bool_env("GWPY_CACHE", False)


@pytest.fixture(scope="module")
@pytest_skip_network_error
def gw150914_hdf5():
    url, = get_event_urls(
        "GW150914",
        version=3,
        detector="L1",
        duration=32,
        sample_rate=4096,
        format="hdf5",
    )
    return download_file(url, cache=GWPY_CACHE)


@pytest_skip_flaky_network
def test_read_hdf5_gwosc(gw150914_hdf5):
    data = TimeSeries.read(
        gw150914_hdf5,
        format="hdf5.gwosc",
    )
    assert data.span == (1126259447, 1126259479)
    assert data.name == "L1:Strain"
    assert data.max().value == pytest.approx(-4.60035111e-20)


@pytest_skip_flaky_network
def test_read_hdf5_gwosc_state(gw150914_hdf5):
    state = StateVector.read(
        gw150914_hdf5,
        format="hdf5.gwosc",
    )
    assert state.name == "L1:DQmask"
    assert state.max().value == 127



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Generating an inspiral range timeseries
#######################################

One standard figure-of-merit for the sensitivity of a gravitational-wave
detector is the distance to which a binary neutron star (BNS) inspiral
with two 1.4 solar mass components would be detected with a signal-to-noise
ratio (SNR) of 8. We can estimate this using
:func:`gwpy.astro.range_timeseries` directly from the strain readout for
a detector.
"""

# %%
# Data access
# -----------
# First, we need to load some data. We can `fetch` the
# `public data <https://gwosc.org/catalog/>`__
# around the GW170817 BNS merger:

from gwpy.timeseries import TimeSeries
h1 = TimeSeries.get("H1", 1187006834, 1187010930)
l1 = TimeSeries.get("L1", 1187006834, 1187010930)

# %%
# Range estimation
# ----------------
# Then, we can measure the inspiral range directly, at 30 second granularity
# with a 4-second FFT length starting at 10 Hz:

from gwpy.astro import range_timeseries
h1range = range_timeseries(h1, 30, fftlength=4, fmin=10)
l1range = range_timeseries(l1, 30, fftlength=4, fmin=10)

# %%
# Visualisation
# -------------
# We can now plot these trends to see the variation in LIGO
# sensitivity over an hour or so surrounding GW170817:

plot = h1range.plot(
    label="LIGO-Hanford",
    color="gwpy:ligo-hanford",
    figsize=(12, 5),
)
ax = plot.gca()
ax.plot(l1range, label="LIGO-Livingston", color="gwpy:ligo-livingston")
ax.set_ylabel("Angle-averaged sensitive distance [Mpc]")
ax.set_title("LIGO sensitivity to BNS around GW170817")
ax.set_epoch(1187008882)  # <- set 0 on plot to GW170817
ax.legend()
plot.show()

# %%
# Note, the extreme dip in LIGO-Livingston's sensitivity near GW170817
# is caused by a loud, transient noise event, see `Phys. Rev. Lett.
# vol. 119, p. 161101 <http://doi.org/10.1103/PhysRevLett.119.161101>`_
# for more information.
